@@ -3,7 +3,8 @@ use warnings;
 package App::DubiousHTTP::TestServer;
 use Scalar::Util 'weaken';
 use Digest::MD5 'md5_base64';
-use App::DubiousHTTP::Tests::Common qw($TRACKHDR ungarble_url);
+use MIME::Base64 'decode_base64';
+use App::DubiousHTTP::Tests::Common qw($TRACKHDR $CLIENTIP ungarble_url);
 
 use IO::Socket::INET;
 my $IOCLASS;
@@ -17,6 +18,18 @@ my %clients;
 my $DEBUG = 0;
 my %trackhdr;
 
+sub _debug {
+    $DEBUG or return;
+    my $msg = shift;
+    $msg = sprintf($msg,@_) if @_;
+    my $time = localtime();
+    $msg =~s{^}{DEBUG: $time }mg;
+    print STDERR $msg."\n";
+}
+
+# close down properly socket etc if user closes program
+$SIG{TERM} = $SIG{INT} = sub { exit(0) };
+
 sub run {
     shift;
     my ($addr,$sslargs,$response) = @_;
@@ -27,7 +40,7 @@ sub run {
 	$sslargs = eval { IO::Socket::SSL::SSL_Context->new( SSL_server => 1, %$sslargs) }
 	    or die "creating SSL context: $@";
     }
-    my $srv = $IOCLASS->new( LocalAddr => $addr, Listen => 10, Reuse => 1 )
+    my $srv = $IOCLASS->new( LocalAddr => $addr, Listen => 10, ReuseAddr => 1 )
 	or die "listen failed: $!";
     $srv->blocking(0);
     $SELECT->handler($srv,0,sub {
@@ -55,7 +68,7 @@ sub delete_client {
 sub add_client {
     my ($cl,$response,$sslctx) = @_;
     my $addr = $cl->sockhost.':'.$cl->sockport;
-    $DEBUG && warn "new client from $addr\n";
+    $DEBUG && _debug("new client from $addr");
 
     $clients{fileno($cl)}{time} = time();
     weaken($clients{fileno($cl)}{fd} = $cl);
@@ -66,13 +79,18 @@ sub add_client {
 
 sub _install_check_https {
     my ($cl,$response,$sslctx) = @_;
-    $DEBUG && warn "add handler for checking https\n";
+    $DEBUG && _debug("add handler for checking https");
     $SELECT->handler($cl,0,sub {
 	my $cl = shift;
 	my $buf;
-	$DEBUG && warn "socket readable - peek\n";
+	$DEBUG && _debug("socket readable - peek");
 	if (!defined recv($cl,$buf,2,MSG_PEEK)) {
-	    $DEBUG && warn "peek failed: $!\n";
+	    $DEBUG && _debug("peek failed: $!");
+	    delete_client($cl);
+	    return;
+	} elsif ($buf eq '') {
+	    # closed immediately
+	    $DEBUG && _debug("client eof after 0 bytes");
 	    delete_client($cl);
 	    return;
 	}
@@ -85,7 +103,7 @@ sub _install_check_https {
 	    SSL_server => 1,
 	    SSL_reuse_ctx => $sslctx
 	)) {
-	    warn "sslify failed: $IO::Socket::SSL::SSL_ERRR";
+	    warn "sslify failed: $IO::Socket::SSL::SSL_ERROR";
 	    delete_client($cl);
 	    return;
 	}
@@ -108,7 +126,7 @@ sub _install_https {
 	} elsif ($IO::Socket::SSL::SSL_ERROR == IO::Socket::SSL::SSL_WANT_WRITE()) {
 	    $SELECT->mask($cl, 0 => 0, 1 => 1);
 	} else {
-	    warn "sslify failed: $IO::Socket::SSL::SSL_ERRR";
+	    warn "sslify failed: $IO::Socket::SSL::SSL_ERROR";
 	    delete_client($cl);
 	    return;
 	}
@@ -127,10 +145,17 @@ sub _install_http {
     my $read = sub {
 	my $cl = shift;
 	my $n = sysread($cl,$rbuf,8192,length($rbuf));
-	$DEBUG && warn "read on ".fileno($cl)." -> ".(defined $n ? $n : $!)."\n";
+	$DEBUG && _debug("read on ".fileno($cl)." -> ".(defined $n ? $n : $!));
 	if ( !$n ) {
 	    # close on eof or error
-	    delete_client($cl) if defined($n) || ! $!{EAGAIN}; 
+	    if (defined($n) || ! $!{EAGAIN}) {
+		if ($clen) {
+		    warn "ERROR: client closed with $clen bytes outstanding";
+		    $payload =~s{^}{DATA|}mg;
+		    print STDERR $payload;
+		}
+		delete_client($cl);
+	    }
 	    return;
 	}
 
@@ -150,17 +175,24 @@ sub _install_http {
 	    return if $clen>0; # need more
 
 	    my $addr = $cl->sockhost.':'.$cl->sockport;
-	    if ( ! eval { $wbuf .= $response->($page,$addr,$hdr,$payload,$ssl) } ) {
+	    if ( ! eval {
+		$CLIENTIP = $cl->peerhost;
+		$CLIENTIP =~s{^::ffff:}{};
+		$wbuf .= $response->($page,$addr,$hdr,$payload,$ssl);
+		$CLIENTIP = undef;
+		1;
+	    } ) {
 		warn "[$page] creating response failed: $@";
 		delete_client($cl);
 		return;
 	    }
-		
+
 	    $clen = $hdr = undef;
 	    if (!$close) {
 		if ($wbuf =~m{(\r?\n)\1}g) {
 		    $close = _mustclose( substr($wbuf,0,pos($wbuf)) );
 		} else {
+		    $DEBUG && _debug("set close=1 because of no header end in wbuf=$wbuf");
 		    $close = 1;
 		}
 	    }
@@ -172,8 +204,49 @@ sub _install_http {
 	    # read header
 	    $hdr = substr($rbuf,0,pos($rbuf),'');
 	    my ($line) = $hdr =~m{^([^\r\n]*)};
-	    $line = ungarble_url($line);
+
+	    my $peer = $cl->peerhost;
+	    $peer =~s{^::ffff:}{};
+
+	    my $urlip;
+	    $line = ungarble_url($line,\$urlip);
 	    $line =~s{\?rand=0\.\d+ }{ };  # remove random for anti-caching
+	    my $ip_mismatch = ($urlip && $urlip ne $peer) ?  "| original($urlip)" : "";
+
+	    (my $method,$page) = $line =~m{ \A
+		(GET|POST) [\040]+
+		(/\S*) [\040]+
+		HTTP/1\.[01] \z
+	    }x or do {
+		warn localtime()." | $peer | badhdr | $line\n";
+		$wbuf .= "HTTP/1.0 204 ok\r\n\r\n";
+		$close = 1;
+		$write->($cl);
+		return;
+	    };
+
+	    if ($page =~m{^/([a-zA-Z0-9_\-]+={0,2})$}) {
+		# maybe base64
+		my $data = $1;
+		$data =~tr{_-}{+/};
+		$data = eval { decode_base64($data) };
+		if (! defined $data) {
+		    warn "base64 decode failed: $@";
+		} elsif ( $data =~m{^(\S+)\0(\d+)\0(.*)\z}s ) {
+		    (my $ref, my $i,$data) = ($1,$2,$3);
+		    my $len = length($data);
+		    $data =~s{\\}{\\\\}g;
+		    $data =~s{\n}{\\n}g;
+		    $data =~s{\r}{\\r}g;
+		    $data =~s{\t}{\\t}g;
+		    printf STDERR "S|%s|%s|%05d|%03d|%s\n",$peer,$ref,$i,$len,$data;
+		    $wbuf .= "HTTP/1.1 200 ok\r\nContent-length: 0\r\n\r\n";
+		    $write->($cl);
+		    return;
+		} else {
+		    #warn "data have not the right format";
+		}
+	    }
 
 	    my $digest = '';
 	    if ($TRACKHDR) {
@@ -194,23 +267,13 @@ sub _install_http {
 		    $xhdr =~s{^}{ |$digest|- }mg;
 		    warn " |$digest|-BEGIN $accept | $ua\n |$digest|- $line\n$xhdr";
 		}
-		warn localtime()." |$digest| ". $cl->peerhost." | $line".($ssl ? " | $ssl":"")."\n";
+		warn localtime()." |$digest| $peer | $line".($ssl ? " | $ssl":"")."$ip_mismatch\n";
 	    } else {
 		my $ua = $hdr =~m{^User-Agent:\s*([^\r\n]+)}mi && $1 || 'Unknown-UA';
 		my @via = $hdr =~m{^Via:\s*([^\r\n]*)}mig;
-		warn localtime()." | $ua  | ". $cl->peerhost." | $line | @via\n";
+		warn localtime()." | $ua  | $peer | $line | @via$ip_mismatch\n";
 	    }
 
-	    (my $method,$page) = $line =~m{ \A 
-		(GET|POST) [\040]+ 
-		(/\S*) [\040]+ 
-		HTTP/1\.[01] \z
-	    }x or do {
-		$wbuf .= "HTTP/1.0 204 ok\r\n\r\n";
-		$close = 1;
-		$write->($cl);
-		return;
-	    };
 	    $clen = $method eq 'POST' && $hdr =~m{^Content-length:[ \t]*(\d+)}mi && $1 || 0;
 	    if ($clen > 2**22) {
 		warn "request body too large ($clen)";
@@ -236,6 +299,7 @@ sub _install_http {
 	    # nothing to write
 	    if ($rbuf eq '' && $close) {
 		# done
+		$DEBUG && _debug("close client because all done and close flag set");
 		delete_client($cl);
 	    } else {
 		$SELECT->mask($cl,1,0);
@@ -243,7 +307,7 @@ sub _install_http {
 	    return;
 	} 
 	my $n = syswrite($cl,$wbuf);
-	$DEBUG && warn "write on ".fileno($cl)." -> ".(defined $n ? $n : $!)."\n";
+	$DEBUG && _debug("write on ".fileno($cl)." -> ".(defined $n ? $n : $!));
 	if ( ! $n ) {
 	    if ( defined($n) || ! $!{EAGAIN} ) {
 		# connection broke
@@ -267,13 +331,22 @@ sub _install_http {
 
 sub _mustclose {
     my $hdr = shift;
-    # keep-alive by header
+    my $close;
+    my $type = $hdr =~m{^[A-Z]+ /} ? 'request':'response';
     while ($hdr =~m{^Connection:[ \t]*(?:(close)|keep-alive)}mig) {
-	return $1 ? 1:0;
+	$close = $1 ? 1: ($close||-1);
     }
-    # keep-alive by response
-    $hdr =~m{\A.* HTTP/1\.(?:0|(1))} or return 1;
-    return $1 ? 0:1;
+    if ($close) {
+	$close = 0 if $close<0;
+	$DEBUG && _debug("set close=$close because of connection header in $type");
+    } elsif ($hdr =~m{\A(?:.* )?HTTP/1\.(?:0|(1))}) {
+	$close = $1 ? 0:1;
+	$DEBUG && _debug("set close=$close because of HTTP version in $type");
+    } else {
+	$close = 1;
+	$DEBUG && _debug("set close=$close because no other information are known in $type");
+    }
+    return $close;
 }
 
 package App::DubiousHTTP::TestServer::Select;
@@ -283,12 +356,13 @@ my $maxfn = 0;
 my @handler;
 my @mask = ('','');
 my @tmpmask;
+*_debug = \&App::DubiousHTTP::TestServer::_debug;
 
 sub new { bless {},shift }
 sub delete {
     my ($self,$cl) = @_;
     defined( my $fn = fileno($cl) ) or die "invalid fd";
-    $DEBUG && warn "remove fd $fn\n";
+    $DEBUG && _debug("remove fd $fn");
     vec($mask[0],$fn,1) = vec($mask[1],$fn,1) = 0;
     vec($tmpmask[0],$fn,1) = vec($tmpmask[1],$fn,1) = 0 if @tmpmask;
     $handler[$fn] = undef;
@@ -306,7 +380,7 @@ sub handler {
 	$sub = [ $sub ] if ref($sub) eq 'CODE';
 	splice(@$sub,1,0,$wcl);
 	$handler[$fn][$rw] = $sub;
-	$DEBUG && warn "add handler($fn,$rw)\n";
+	$DEBUG && _debug("add handler($fn,$rw)");
     }
 }
 
@@ -314,7 +388,7 @@ sub mask {
     my ($self,$cl,%val) = @_;
     defined( my $fn = fileno($cl) ) or die "invalid fd";
     while (my ($rw,$val) = each %val) {
-	$DEBUG && warn "set mask($fn,$rw) to $val\n";
+	$DEBUG && _debug("set mask($fn,$rw) to $val");
 	vec($mask[$rw],$fn,1) = $val;
     }
 }
@@ -323,12 +397,12 @@ sub loop {
     loop:
     @tmpmask = @mask;
     my $rv = select($tmpmask[0],$tmpmask[1],undef,undef);
-    $DEBUG && warn "select -> $rv\n";
+    $DEBUG && _debug("select -> $rv");
     die "loop failed: $!" if $rv <= 0;
     for my $rw (0,1) {
 	for( my $fn=0; $fn<=$maxfn; $fn++) {
 	    vec($tmpmask[$rw],$fn,1) or next;
-	    $DEBUG && warn "selected($fn,$rw)\n";
+	    $DEBUG && _debug("selected($fn,$rw)");
 	    my $sub = $handler[$fn][$rw] or die "no handler";
 	    $sub->[0](@{$sub}[1..$#$sub]);
 	}
