@@ -37,43 +37,42 @@ More information about bypassing firewalls using interpretation differences can
 be found <a href="http://noxxi.de/research/semantic-gap.html">here</a>.
 </p>
 
-<h2>Bulk test with innocent payload</h2>
-
-<p>
-This bulk test automatically triggers various kinds of strange HTTP responses
-which contain an innocent payload and compares the payload it gets in the
-browser against the expected payload. It uses XMLHttpRequests for this purpose
-which often but not in all cases show the same behavior as other HTTP requests
-by the browser (i.e. loading image, script,...). 
-In lots of cases the browser will extract the original payload from the response
-even the response itself was invalid. On the other hand there are cases were the
-browser does not get the expected payload even if the response was valid,
-because either the browser or a proxy/firewall in between does not fully
-understand legit HTTP or blocked legit but uncommon HTTP for security reasons.
-</p>
-<p class=runtest><a href="/auto/all/novirus.txt">Run Test</a></p>
-
 <h2>Bulk test with virus payload</h2>
 
 <p>
-This is the same bulk test as the previous test with the exception that the
-payload is a virus this time. The payload consists of the <a
-href="http://www.eicar.org/86-0-Intended-use.html">EICAR test virus</a> which is
-commonly used for basic tests of antivirus and which should be detected by every
-firewall which does deep inspection to filter out malware. 
+This bulk test tries to transfer the <a
+href="http://www.eicar.org/86-0-Intended-use.html">EICAR test virus</a> from the
+server to the client. This test virus is commonly used for basic tests of
+antivirus and should be detected by every firewall which does deep
+inspection to filter out malware. Since this virus itself is not malicious it is
+safe to run this test.
+</p><p>
+But, the transfer is done with various kinds of uncommon or even invalid HTTP
+responses to check if the inspection of the firewall can be bypassed this way.
+The response from the server will then compared to the expected payload and
+hopefully all transfers will be blocked either by the firewall or are considered
+invalid by the browser.
+</p><p>
+The test uses XMLHttpRequests to issue the request and get the response. In most but
+not all cases this shows the same behavior as other HTTP requests by the browser
+(i.e. loading image, script,...). But to verify that an evasion is actually
+possible with normal download one should use the provided link to actually test
+the evasion.
 </p>
+<p id=test_virus class=runtest><a href="/auto/all/eicar.txt">Run Test with <strong>EICAR test virus</strong> payload</a></p>
+
+<h2>Bulk test with innocent payload</h2>
+
 <p>
-The goal of this test is to find out if the firewall interprets the HTTP
-response in a different way then the browser and if this would allow a critical
-bypass of the firewalls protection. Since the EICAR test virus used in this test
-is not malicious it is safe to run this test even if the firewall gets
-successfully bypassed.
-It is important to consider that the XMLHttpRequests used for this tests do 
-behave the same as normal download links in most but not all cases. This means
-to verify that an evasion is actually possible with a download link one should
-use the provided link to actually test the evasion.
+This is the same bulk test as the previous one but this time the payload is
+completely innocent. This test can be used to find out the behavior of the
+browsers itself, i.e. how uncommon or invalid HTTP responses are handled by the
+browser. It can also be used to check if the use of proxies changes this
+behavior and if firewalls block innocent payload if it is transferred using an
+uncommon or invalid HTTP response.
 </p>
-<p class=runtest><a href="/auto/all/eicar.txt">Run Test</a></p>
+<p id=test_novirus class=runtest><a href="/auto/all/novirus.txt">Run Test with <strong>innocent</strong> payload</a></p>
+
 
 <h2>Non-Bulk tests</h2>
 
@@ -122,13 +121,12 @@ sub auto {
     my ($hdr,$body,$isbad) = content($page);
     $html .= "<script>\n";
 
-    $html .= sprintf("reference='%x%x';\n", rand(2**32), time());
-
     my ($accept) = $rqhdr =~m{^Accept:[ \t]*([^\r\n]+)}mi;
     if ($qstring =~m{(?:^|\&)accept=([^&]*)}) {
 	($accept = $1) =~s{(?:%([a-f\d]{2})|(\+))}{ $2 ? ' ' : chr(hex($1)) }esg;
     }
     $html .= "accept = '".quotemeta($accept)."';\n" if $accept;
+    $html .= "fast_feedback = 16384;\n" if $FAST_FEEDBACK;
 
     $html .= "expect64 = '".encode_base64($body,'')."';\n";
     $html .= 'results = "V | '.App::DubiousHTTP->VERSION.'\n";' . "\n";
@@ -158,6 +156,7 @@ sub auto {
 	}
 
     }
+    $html .= sprintf("reference='%x' + Math.floor(time()/1000).toString(16);\n", rand(2**32));
     $html .= "runtests(checks,0);\n</script>\n";
     return "HTTP/1.0 200 ok\r\n".
 	"Content-type: text/html\r\n".
@@ -174,6 +173,7 @@ body      { font-family: Verdana, sans-serif; }
 #nobad    { padding: 2em; margin: 1em; background: #ff3333; display: none; }
 #nobad div   { font-size: 150%; margin: 0.5em;  }
 #noevade  { padding: 1em; margin: 1em; background: green; display: none; }
+#overblock { padding: 1em; margin: 1em; background: #ff9933; display: none; }
 #evadable { padding: 1em; margin: 1em; background: #ff3333; display: none; }
 #urlblock { padding: 1em; margin: 1em; background: #ffff00  ; display: none; }
 #urlblock div  { font-size: 150%; margin: 0.5em;  }
@@ -194,15 +194,21 @@ You need to have JavaScript enabled to run this tests.
 <div id=evasions></td>
 <div id=process></div>
 <div id=evadable> </div>
+<div id=overblock> </div>
 <div id=noevade> </div>
 <div id=warnings><h1>Serious Problems</h1><ol id=ol_warnings></ol></div>
 <div id=notice><h1>Behavior in Uncommon Cases</h1><ol id=ol_notice></ol></div>
 <div id=debug><h1>Debug</h1></div>
 <script>
-function base64_encode(input) {
+
+var time = Date.now || function() { return +new Date; };
+
+function base64_encode(input,urlsafe) {
     var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
     var output = "";
-    var chrs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
+    var chrs = urlsafe
+	? "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_="
+	: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
 
     for(var i=0;i<input.length; i+=3) {
 	chr1 = input.charCodeAt(i) & 0xff;
@@ -230,6 +236,28 @@ function base64_encode(input) {
     return output;
 }
 
+function base64_decode(input) {
+    input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");
+    var chrs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
+
+    var chr = new Array(3);
+    var enc = new Array(4);
+    var output = "";
+    for(var i=0;i<input.length;i+=4) {
+	for(var j=0;j<4;j++) {
+	    enc[j] = chrs.indexOf(input.charAt(i+j));
+	}
+
+        output = output + 
+            String.fromCharCode( (enc[0] << 2) | (enc[1] >> 4) )
+            + ((enc[2] != 64) ? String.fromCharCode( ((enc[1] & 15) << 4) | (enc[2] >> 2) ) : '')
+            + ((enc[3] != 64) ? String.fromCharCode( ((enc[2] & 3) << 6) | enc[3] ) : '');
+    }
+
+    return output;
+}
+
+
 
 var div_debug = document.getElementById('debug');
 var div_notice = document.getElementById('notice');
@@ -248,7 +276,7 @@ var accept = null;
 
 function add_warning(m,test) {
     div_ol_warnings.innerHTML = div_ol_warnings.innerHTML + "<li>" + m + ": <span class=desc>" + test['desc'] + "</span>" +
-	"&nbsp;<a class=trylink download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
+	"&nbsp;<a class=trylink target=_blank download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
 	"&nbsp;<a class=srclink target=_blank href=/src" + test['page'] + ">src</a>" +
 	"</li>";
     div_warnings.style.display = 'block';
@@ -256,7 +284,7 @@ function add_warning(m,test) {
 
 function add_notice(m,test) {
     div_ol_notice.innerHTML = div_ol_notice.innerHTML + "<li>" + m + ": <span class=desc>" + test['desc'] + "</span>" +
-	"&nbsp;<a class=trylink download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
+	"&nbsp;<a class=trylink target=_blank download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
 	"&nbsp;<a class=srclink target=_blank href=/src" + test['page'] + ">src</a>" +
 	"</li>";
     div_notice.style.display = 'block';
@@ -264,7 +292,7 @@ function add_notice(m,test) {
 
 function add_debug(m,test) {
     div_debug.innerHTML = div_debug.innerHTML + m + (test ?
-	"&nbsp;<a class=trylink download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
+	"&nbsp;<a class=trylink target=_blank download='" + test['file'] + "' href=" + test['page'] + ">try</a>" +
 	"&nbsp;<a class=srclink target=_blank href=/src" + test['page'] + ">src</a>" +
 	"<br>": "");
 }
@@ -285,6 +313,8 @@ function _log(m) {
 
 var evasions = 0;
 var evasions_blocked = 0;
+var overblocked = 0;
+var maybe_overblocked = 0;
 var browser_invalid = 0;
 function xhr(method,page,payload,callback) {
     var req = null;
@@ -359,7 +389,14 @@ function check_page(req,test,status) {
 	    if (result64 == expect) {
 		status = 'match';
 	    } else {
-		status = 'change(' + status + ')';
+		var pn = response.indexOf(base64_decode(expect));
+		if (pn<0) {
+		    status = 'change(' + status + ')';
+		} else {
+		    _log( 'off=' + pn + 'response="' + response + '" expect="' + base64_decode(expect) + '"');
+		    status = 'change(' + status + ',off=' + pn + ')';
+		}
+		    
 		results = results + "R | " + test['page'] + " | " + response.length + " | " + base64_encode((header + "--\n" + response).substr(0,1000)) + "\n";
 		_log( "len=" + response.length + "   " + test['page'] + ' - ' + test['desc'] );
 		_log( "response: " + result64 );
@@ -380,8 +417,9 @@ function check_page(req,test,status) {
 	if (status != 'match') {
 	    browser_invalid++;
 	    if (test['valid'] == 2) { // no browser should fail on this!
+		overblocked++;
 		add_warning("Failed to load harmless and perfectly valid response",test);
-		results = results + "X | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless\n";
+		results = results + "X | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless but must succeed\n";
 		results = results + "T | " + test['page'] + " | " + result64 + "\n";
 		var div_urlblock = document.getElementById('urlblock');
 		div_urlblock.innerHTML = "<div>" 
@@ -391,9 +429,16 @@ function check_page(req,test,status) {
 		    + "the abilities of the firewall to detect malware."
 		    + "</div>";
 		div_urlblock.style.display = 'block';
-	    } else if (['valid']>0) {
-		add_notice("Failed to load harmless and valid response, might be browser bug",test);
-		results = results + "X | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless\n";
+	    } else if (test['valid']>0) {
+		if (test['valid']==3) {
+		    // firewall might have modified request 
+		    maybe_overblocked++;
+		    add_warning("Failed to load harmless and valid response, maybe the firewall blocked too much",test);
+		    results = results + "X | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless but should succeed\n";
+		} else {
+		    add_notice("Failed to load harmless and valid response, might be browser bug",test);
+		    results = results + "X | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless\n";
+		}
 		results = results + "T | " + test['page'] + " | " + result64 + "\n";
 	    } else {
 		results = results + "B | " + status + " | " + test['page'] + " | " + test['desc'] + " | failed harmless\n";
@@ -485,20 +530,37 @@ function runtests(todo,done) {
 		todo.unshift({ page: test['harmless_page'], desc: test['desc'], valid: test['valid'], harmless_retry:1,
 		    retry4status:status, retry4page: test['page']});
 	    }
+	    if (fast_feedback && results.length > fast_feedback) {
+		submit_part();
+	    }
 	    runtests(todo,done+1);
 	});
     } else {
 	div_process.style.display = 'none';
 	add_debug("*DONE*");
+	var submit_url;
 	if (isbad) {
 	    var div;
-	    if (evasions == 0) {
+	    if (evasions == 0 && overblocked == 0) {
 		results = results + "NO EVASIONS\n";
 		div = document.getElementById('noevade');
 		div.innerHTML = "<h1>Congratulations!<br>No evasions detected.</h1>"
 		    + evasions_blocked + " evasions attempts were blocked by the firewall and " 
 		    + browser_invalid + " attempts failed because the browser considered the response invalid or because the firewall blocks (invalid) responses even if there is no malware payload."
 		    + "Please note that these might be considered valid by other browsers and might lead to possible evasions, so better try with other browsers too."
+		    + "For this reason I would recommend to check with at least Firefox, Chrome, Safari, Internet Explorer, Edge and Opera because they all behave differently."
+		    + "<br><br>To get an overview which products behave that nicely "
+		    + "it would be helpful if you provide us with information about the firewall product you use. "
+		    + "Please add as much details as you know and like to offer, i.e. model, patch level, specific configurations. ";
+	    } else if (evasions == 0) {
+		results = results + "NO EVASIONS BUT OVERBLOCKING\n";
+		div = document.getElementById('overblock');
+		div.innerHTML = "<h1>Suspicious!<br>No evasions detected but it looks like overblocking.</h1>"
+		    + evasions_blocked + " evasions attempts were blocked by the firewall but in at least " 
+		    + overblocked + " cases the firewall blocked perfectly valid and innocent responses."
+		    + browser_invalid + " attempts failed because the browser considered the response invalid or because the firewall blocks (invalid) responses even if there is no malware payload."
+		    + "Please note that these might be considered valid by other browsers and might lead to possible evasions, so better try with other browsers too."
+		    + "For this reason I would recommend to check with at least Firefox, Chrome, Safari, Internet Explorer, Edge and Opera because they all behave differently."
 		    + "<br><br>To get an overview which products behave that nicely "
 		    + "it would be helpful if you provide us with information about the firewall product you use. "
 		    + "Please add as much details as you know and like to offer, i.e. model, patch level, specific configurations. ";
@@ -506,9 +568,11 @@ function runtests(todo,done) {
 		div = document.getElementById('evadable');
 		div.innerHTML = "<h1>Danger!<br>Possible evasions detected!</h1>"
 		    + "The test detected that " + evasions + " evasion attempts were not blocked by the firewall.<br>"
+		    + ((overblocked>0) ? "Additionally in " + overblocked + " cases the firewall blocked perfectly valid and innocent responses.<br>" : '' )
 		    + evasions_blocked + " evasions attempts were blocked by the firewall and " 
 		    + browser_invalid + " attempts failed because the browser considered the response invalid or because the firewall blocks (invalid) responses even if there is no malware payload."
-		    + "Please note that these might be considered valid by other browsers and might lead to possible evasions, so better try with other browsers too.<br>"
+		    + "Please note that these might be considered valid by other browsers and might lead to possible evasions, so better try with other browsers too."
+		    + "For this reason I would recommend to check with at least Firefox, Chrome, Safari, Internet Explorer, Edge and Opera because they all behave differently.<br>"
 		    + "Since the test differs slightly from a manually triggered download it might be that some of the detected evasions are "
 		    + "not usable in reality, so please make sure the evasion works by clicking the [TRY] link "
 		    + "and comparing the downloaded file with the EICAR test virus. The file should be 68 byte and contain the string "
@@ -518,17 +582,73 @@ function runtests(todo,done) {
 		    + "Please add as much details as you know and like to offer, i.e. model, patch level, specific configurations. ";
 	    }
 	    div.innerHTML += '<br><br><form enctype="multipart/form-data" method=POST action="/submit_details/' + reference + '/evasions=' + evasions + "/evasions_blocked=" + evasions_blocked + '">'
-		+ '<input type=hidden name=results value="' + escapeAttribute(results) + '">'
 		+ '<textarea name=product cols=80 rows=4>... please add product description here ...</textarea>'
 		+ '<br><input type=submit name=Send></form>';
 	    div.style.display = 'block';
-	    xhr('POST','/submit_results/' + reference + '/evasions=' + evasions + "/evasions_blocked=" + evasions_blocked, results, null);
+	    submit_url = '/submit_results/' + reference + '/evasions=' + evasions + "/evasions_blocked=" + evasions_blocked;
 	} else {
-	    xhr('POST','/submit_results/' + reference ,results, null);
+	    submit_url = '/submit_results/' + reference;
+	}
+
+	if (submit_url) {
+	    submit_result(submit_url,results);
+	    results = null;
 	}
     }
 }
 
+function submit_result(url,data) {
+    xhr('POST', url, data, function(req) {
+	var blocked = 1;
+	try {
+	    if (req.status != 200) {
+		_log("bad status from submit: " + req.status);
+	    } else if (req.getResponseHeader("X-ID") != url) {
+	    	_log("bad response x-id:'" + req.getResponseHeader("X-ID") + "' expect:'" + url +"'");
+	    } else {
+		_log("submission ok");
+		blocked = 0
+	    }
+	}
+	catch(e) { _log(e); }
+
+	// disable - to much false reports. Better use --fast-feedback
+	if ( 0 && blocked) {
+	    // POST might be blocked, try as lots of GET requests
+	    var post = url + "\n" + data;
+	    data = null;
+	    var i = 0;
+	    var submit_part;
+	    submit_part = function(req,status) {
+		if (!status) status = req.status;
+		if (status != 200) {
+		    _log("submitting part failed: " + status);
+		    return;
+		} else if (post == null) {
+		    return;
+		}
+
+		var buf = post.substr(0,512);
+		if (buf.length<512) {
+		    post = null;
+		} else {
+		    post = post.substr(512);
+		}
+		buf = base64_encode( reference + "\0" + i + "\0" + buf,1);
+		i++;
+		xhr('GET', '/' + buf, null, submit_part);
+	    };
+	    submit_part(null,200);
+	}
+    });
+}
+
+var partid = 0;
+function submit_part() {
+    submit_result("/submit_part/" + reference + "/" + partid++,results);
+    results = '';
+}
+
 document.getElementById('noscript').style.display = 'none';
 </script>
 HTML
