@@ -19,20 +19,41 @@ DESC
 
     # ------------------------ Tests -----------------------------------
     [ 'VALID: basic tests' ],
-    [ VALID, 'chunked' => 'simple and valid chunking'],
-    [ VALID, 'clen' => 'content-length header, not chunked'],
+    [ SHOULDBE_VALID, 'chunked' => 'simple and valid chunking'],
+    [ MUSTBE_VALID, 'clen' => 'content-length header, not chunked'],
 
-    [ 'VALID: use of extensions in chunked header' ],
-    [ VALID, 'chunked-ext-junk' => "chunked with some junk chunk extension" ],
-    [ VALID, 'chunked-ext-chunk' => "chunked with some junk chunk extension looking like a chunk" ],
+    [ 'VALID: modification of chunk size' ],
+    [ UNCOMMON_VALID, '0size' => "chunks size prefixed with 0" ],
+    [ UNCOMMON_VALID, '00size' => "chunks size prefixed with 00" ],
+    [ UNCOMMON_VALID, 'ucsize' => "upper case size" ],
+    [ UNCOMMON_VALID, '0ucsize' => "upper case size prefix with 0" ],
+    [ INVALID, 'size-space' => "size followed by space" ],
+    [ INVALID, 'size-tab' => "size followed by tab" ],
+    [ INVALID, 'size-cr' => "size followed by <CR>" ],
+    [ INVALID, 'size-lf' => "size followed by <LF>" ],
+    [ INVALID, 'size-x' => "size followed by char 'x'" ],
+    [ INVALID, 'size-spacex' => "size followed by space and char 'x'" ],
+    [ INVALID, 'space-size' => "size prefixed by space" ],
+    [ INVALID, 'tab-size' => "size prefixed by tab" ],
+    [ INVALID, 'cr-size' => "size prefixed by cr" ],
+    [ INVALID, 'x-size' => "size prefixed by char 'x'" ],
+    [ INVALID, 'xspace-size' => "size prefixed by char 'x' and space" ],
+    [ UNCOMMON_VALID, 'final=00' => 'final chunk size "00"' ],
+    [ UNCOMMON_VALID, 'final=00000000000000000000' => 'final chunk size "00000000000000000000"' ],
+    [ INVALID, 'final=0x' => 'final chunk size "0x"' ],
+
+    [ 'VALID: (but uncommon) use of extensions in chunked header' ],
+    [ UNCOMMON_VALID, 'chunk-ext-junk' => "chunked with some junk chunk extension" ],
+    [ UNCOMMON_VALID, 'chunk-ext-chunk' => "chunked with some junk chunk extension looking like a chunk" ],
 
     [ 'VALID: combined with content-length' ],
     # according to RFC2616 TE chunked has preference to clen
     [ VALID, 'chunked,clen' => 'chunked first then content-length, served chunked'],
     [ VALID, 'clen,chunked' => 'content-length first then chunked, served chunked'],
     # but some still expect clen bytes
-    [ VALID, 'chunked,clen200' => 'chunking and content-length header with double length, served chunked'],
-    [ VALID, 'chunked,clen50'  => 'chunking and content-length header with half length, served chunked'],
+    # safari does not like it, so mark it as uncommon
+    [ UNCOMMON_VALID, 'chunked,clen200' => 'chunking and content-length header with double length, served chunked'],
+    [ UNCOMMON_VALID, 'chunked,clen50'  => 'chunking and content-length header with half length, served chunked'],
 
     [ 'INVALID: chunking is only allowed with HTTP/1.1' ],
     [ INVALID, 'chunked,http10' => 'Chunked Header and HTTP/1.0. Served chunked.'],
@@ -65,7 +86,9 @@ DESC
     [ 'VALID: valid variations on "chunked" value' ],
     [ VALID, 'chUnked' => 'mixed case "chUnked", served chunked'],
     [ UNCOMMON_VALID,'nl-chunked' => "chunked header with continuation line, served chunked"],
+    [ UNCOMMON_VALID,'chunkednl-' => "chunked header followed by empty with continuation line, served chunked"],
     [ UNCOMMON_VALID,'nl-nl-chunked' => "chunked header with double continuation line, served chunked"],
+    [ UNCOMMON_VALID,'nl-nl-chunked-nl-' => "chunked header with double continuation line and continuation afterwareds, served chunked"],
 
     [ 'INVALID: invalid variations on "chunked" value' ],
     [ INVALID, 'chu' => '"chu" not "chunked"'],
@@ -77,11 +100,19 @@ DESC
     [ UNCOMMON_INVALID, 'chunked-x,do_clen' => '"chunked x" not "chunked", not served chunked'],
     [ UNCOMMON_INVALID, 'x-chunked,do_clen' => '"x chunked" not "chunked", not served chunked'],
     [ INVALID, 'x-nl-chunked' => '"x-folding-chunked" not "chunked"'],
+    [ INVALID, 'chunked-nl-x' => '"chunked-folding-x" not "chunked"'],
     [ INVALID, 'rfc2047,do_chunked' => 'rfc2047/base64 encoded "chunked", served chunked' ],
     [ UNCOMMON_VALID, 'rfc2047,do_clen' => 'rfc2047/base64 encoded "chunked", not served chunked' ],
     [ UNCOMMON_VALID, 'rfc2047,clen,do_clen' => 'rfc2047/base64 encoded "chunked" and content-length, not served chunked' ],
     [ INVALID,'nl-chunked,do_clen' => "chunked header with continuation line. Not served chunked."],
+    [ INVALID,'chunkednl-,do_clen' => "chunked header followed by empty continuation line. Not served chunked."],
     [ INVALID,'nl-nl-chunked,do_clen' => "chunked header with double continuation line, not served chunked"],
+    [ INVALID,'crchunked,do_chunked' => "Transfer-Encoding:<CR>chunked. Served chunked."],
+    [ INVALID,'crchunked,do_clen' => "Transfer-Encoding:<CR>chunked. Not served chunked."],
+    [ INVALID,'cr-chunked,do_chunked' => "Transfer-Encoding:<CR><space>chunked. Served chunked."],
+    [ INVALID,'cr-chunked,do_clen' => "Transfer-Encoding:<CR><space>chunked. Not served chunked."],
+    [ INVALID,'chunkedcr-,do_chunked' => "Transfer-Encoding:chunked<CR><space>. Served chunked."],
+    [ INVALID,'chunkedcr-,do_clen' => "Transfer-Encoding:chunked<CR><space>. Not served chunked."],
     [ INVALID,'ce-chunked,do_chunked' => "Content-encoding chunked instead of Transfer-encoding. Served chunked."],
 
     [ 'INVALID: hiding with another Transfer-Encoding' ],
@@ -99,7 +130,11 @@ DESC
 
     [ 'INVALID: hiding the Transfer-Encoding header' ],
     [ INVALID, 'space-colon-chunked,do_chunked' => '"Transfer-Encoding<space>:", served chunked' ],
+    [ INVALID, 'tab-colon-chunked,do_chunked' => '"Transfer-Encoding<tab>:", served chunked' ],
+    [ INVALID, 'cr-colon-chunked,do_chunked' => '"Transfer-Encoding<CR>:", served chunked' ],
     [ UNCOMMON_INVALID, 'space-colon-chunked,do_clen' => '"Transfer-Encoding<space>:", not served chunked' ],
+    [ UNCOMMON_INVALID, 'tab-colon-chunked,do_clen' => '"Transfer-Encoding<tab>:", not served chunked' ],
+    [ UNCOMMON_INVALID, 'cr-colon-chunked,do_clen' => '"Transfer-Encoding<CR>:", not served chunked' ],
     [ INVALID, 'colon-colon-chunked,do_chunked' => '"Transfer-Encoding::", served chunked' ],
     [ UNCOMMON_INVALID, 'colon-colon-chunked,do_clen' => '"Transfer-Encoding::", not served chunked' ],
     [ INVALID, 'cronly-chunked,do_chunked' => 'Transfer-Encoding with only <CR> as line delimiter before, served chunked' ],
@@ -108,7 +143,17 @@ DESC
     [ INVALID, 'lfonly-chunked,do_clen' => 'Transfer-Encoding with only <LF> as line delimiter before, not served chunked' ],
 
     [ 'INVALID: invalid chunks' ],
-    [ INVALID, 'chunked-lf' => "chunk with LF as delimiter instead of CRLF" ],
+    [ INVALID, 'chunk-lf' => "chunk with LF as delimiter instead of CRLF" ],
+    [ INVALID, 'chunk-cr' => "chunk with CR as delimiter instead of CRLF" ],
+    [ INVALID, 'chunk-crcr' => "chunk with CRCR as delimiter instead of CRLF" ],
+    [ INVALID, 'chunk-lflf' => "chunk with LFLF as delimiter instead of CRLF" ],
+    [ INVALID, 'chunk-lfcr' => "chunk with LFCR as delimiter instead of CRLF" ],
+    [ INVALID, 'nofinal' => 'missing final chunk' ],
+    [ INVALID, 'eof-inchunk' => 'eof inside some chunk' ],
+    [ INVALID, 'space-before-chunks' => 'space before chunks start' ],
+    [ INVALID, 'lf-before-chunks' => '<LF> before chunks start' ],
+    [ INVALID, 'cr-before-chunks' => '<CR> before chunks start' ],
+    [ INVALID, 'crlf-before-chunks' => '<CR><LF> before chunks start' ],
 );
 
 
@@ -117,38 +162,37 @@ sub make_response {
     return make_index_page() if $page eq '';
     my ($hdr,$data) = content($page,$spec) or die "unknown page $page";
     my $version = 'HTTP/1.1';
-    my $te;
+    my ($te,@chunks,%chunkmod);
+    my $sizefmt = '%x';
+    my $before_chunks = '';
+    my $final = '0';
     for (split(',',$spec)) {
-	if ( ! $_ || $_ eq 'chunked' ) {
-	    $hdr .= "Transfer-Encoding: chunked\r\n"
-	} elsif ( $_ eq 'space-colon-chunked' ) {
+	if ( m{^(x|-|nl|lf|cr)*chunked(x|-|nl|lf|cr)*$}i ) {
+	    s{-}{ }g;
+	    s{nl}{\r\n}g;
+	    s{lf}{\n}g;
+	    s{cr}{\r}g;
+	    $hdr .= "Transfer-Encoding: $_\r\nConnection: close\r\n";
+	} elsif ( m{^(space|tab|cr|colon)-colon-chunked$} ) {
+	    my $c = $1;
+	    $c =~s{space}{ }g;
+	    $c =~s{colon}{:}g;
+	    $c =~s{tab}{\t}g;
+	    $c =~s{cr}{\r}g;
 	    $te = 'chunked';
-	    $hdr .= "Transfer-Encoding : chunked\r\n"
-	} elsif ( $_ eq 'colon-colon-chunked' ) {
-	    $te = 'chunked';
-	    $hdr .= "Transfer-Encoding:: chunked\r\n"
+	    $hdr .= "Connection: close\r\nTransfer-Encoding$c: chunked\r\n"
 	} elsif ( my ($crlf) = m {^(lf|cr)only-chunked$} ) {
 	    $te = 'chunked';
 	    $hdr = "X-Foo: bar" if $hdr !~s{\r\n\z}{};
 	    $hdr .= ($crlf eq 'lf') ? "\n":"\r";
-	    $hdr .= "Transfer-Encoding: chunked\r\n";
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
 	} elsif ( $_ eq '1chunk' ) {
 	    $hdr .= "Transfer-Encoding: chunked\r\n";
-	    $te = $_
-	} elsif ( $_ eq 'chUnked' ) {
-	    $hdr .= "Transfer-Encoding: chUnked\r\n"
+	    @chunks = $data;
 	} elsif ( $_ eq 'chu' ) {
-	    $hdr .= "Transfer-Encoding: chu\r\n"
+	    $hdr .= "Transfer-Encoding: chu\r\nConnection: close\r\n"
 	} elsif ( $_ eq 'ce-chunked' ) {
-	    $hdr .= "Content-Encoding: chunked\r\n"
-	} elsif ( $_ eq 'xchunked' ) {
-	    $hdr .= "Transfer-Encoding: xchunked\r\n"
-	} elsif ( $_ eq 'chunkedx' ) {
-	    $hdr .= "Transfer-Encoding: chunkedx\r\n"
-	} elsif ( $_ eq 'x-chunked' ) {
-	    $hdr .= "Transfer-Encoding: x chunked\r\n"
-	} elsif ( $_ eq 'chunked-x' ) {
-	    $hdr .= "Transfer-Encoding: chunked x\r\n"
+	    $hdr .= "Content-Encoding: chunked\r\nConnection: close\r\n"
 	} elsif ( $_ =~ m{^clen(\d+)?$} ) {
 	    $hdr .= "Content-length: ". int(($1||100)/100*length($data)) ."\r\n"
 	} elsif ( $_ eq 'http10' ) {
@@ -160,55 +204,77 @@ sub make_response {
 	} elsif ( $_ eq 'do_chunked' ) {
 	    $te = 'chunked'
 	} elsif ( $_ eq 'chunked-semicolon' ) {
-	    $hdr .= "Transfer-Encoding: chunked;\r\n"
-	} elsif ( m{^chunked-ext|^chunked-lf}) {
-	    $hdr .= "Transfer-Encoding: chunked\r\n";
-	    $te = $_
+	    $hdr .= "Transfer-Encoding: chunked;\r\nConnection: close\r\n"
 	} elsif ( $_ eq 'rfc2047' ) {
-	    $hdr .= "Transfer-Encoding: =?UTF-8?B?Y2h1bmtlZAo=?=\r\n";
+	    $hdr .= "Transfer-Encoding: =?UTF-8?B?Y2h1bmtlZAo=?=\r\nConnection: close\r\n";
 	} elsif ( $_ eq 'xte' ) {
-	    $hdr .= "Transfer-Encoding: lalala\r\n";
-	} elsif ( m{^(.*-)chunked$} ) {
-	    my $prefix = $1;
-	    $prefix =~s{-}{ }g;
-	    $prefix =~s{\bnl\b}{\r\n}g;
-	    $hdr .= "Transfer-Encoding: ${prefix}chunked\r\n";
+	    $hdr .= "Transfer-Encoding: lalala\r\nConnection: close\r\n";
+	} elsif ( m{^(chunk-ext-|nofinal$|eof-inchunk$)} ) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    $chunkmod{$_} = 1;
+	} elsif ( my ($eol) = m{^chunk-((?:lf|cr)+)$} ) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    $eol =~s{cr}{\r}g;
+	    $eol =~s{lf}{\n}g;
+	    $chunkmod{lineend} = $eol;
+	} elsif ( m{^(-|space|cr|lf|tab|x)*(0*)(uc)?size(-|space|cr|lf|tab|x)*$}) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    @chunks = ( $data =~m{(.{1,15})}smg,'') if ! @chunks;
+	    s{ucsize}{%X};
+	    s{size}{%x};
+	    s{\\r}{\r}g;
+	    s{\\n}{\n}g;
+	    s{-}{}g;
+	    s{space}{ }g;
+	    s{tab}{\t}g;
+	    s{cr}{\r}g;
+	    s{lf}{\n}g;
+	    $sizefmt = $_;
+	} elsif (m{((?:space|tab|cr|lf)*)-before-chunks}) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    $before_chunks = $1;
+	    $before_chunks =~ s{space}{ }g;
+	    $before_chunks =~ s{tab}{\t}g;
+	    $before_chunks =~ s{cr}{\r}g;
+	    $before_chunks =~ s{lf}{\n}g;
+	} elsif (m{^final=(.*)$}) {
+	    $hdr .= "Transfer-Encoding: chunked\r\nConnection: close\r\n";
+	    $final = $1;
 	} else {
 	    die $_
 	}
     }
     $hdr = "$version 200 ok\r\n$hdr";
     $te ||= $hdr =~m{^Transfer-Encoding:}im ? 'chunked':'clen';
-    if ( $te eq 'chunked' ) {
-	$data = join("", 
-	    map { sprintf("%x\r\n%s\r\n",length($_),$_) } 
-	    ( $data =~m{(..?)}smg,'')
-	)
-    } elsif ( $te eq '1chunk' ) {
-	$data = sprintf("%x\r\n%s\r\n0\r\n\r\n",length($data),$data);
-    } elsif ($te eq 'chunked-ext-junk') {
-	$data = join("", 
-	    map { sprintf("%x; foobar\r\n%s\r\n",length($_),$_) } 
-	    ( $data =~m{(..?)}smg,'')
-	)
-    } elsif ($te eq 'chunked-ext-chunk') {
-	$data = join("", 
-	    map { sprintf(
-		"%x; %s  %x\r\n%s\r\n",
-		length($_),                    # chunk length
-		"x" x length($_), length($_),  # chunk extensions looking like chunk if any two bytes are skipped instead of \r\n
-		$_                             # chunk
-	    )} 
-	    ( $data =~m{(..?)}smg,'')
-	)
-    } elsif ($te eq 'chunked-lf') {
-	$data = join("", 
-	    map { sprintf("%x\n%s\n",length($_),$_) } 
-	    ( $data =~m{(..?)}smg,'')
-	)
-    }
+    @chunks = ( $data =~m{(.{1,5})}smg,'') if $te eq 'chunked' && ! @chunks;
+    if (@chunks) {
+	@chunks = map { [ length($_), $_ ] } @chunks;
+	my $nl = $chunkmod{lineend} || "\r\n";
+	if ($chunkmod{'chunk-ext-chunk'}) {
+	    $_->[2] = sprintf("; %s  %x","x" x $_->[0],$_->[0]) for @chunks;
+	} elsif ($chunkmod{'chunk-ext-junk'}) {
+	    $_->[2] = "; foobar" for @chunks;
+	}
+	pop @chunks if $chunkmod{nofinal} && ! $chunks[-1][0];
 
-    return "$hdr\r\n$data";
+	my $end = '';
+	if ($chunkmod{'eof-inchunk'}) {
+	    pop @chunks if ! $chunks[-1][0]; # remove final chunk
+	    my $last = pop(@chunks);
+	    $end = sprintf("%x%s%s%s",$last->[0]+10,$last->[2]||'',$nl,$last->[1]);
+	}
+
+	$data = join("",map { 
+	    $_->[0] ? sprintf("$sizefmt%s%s%s%s",
+		$_->[0],          # size
+		$_->[2] || '',    # ext
+		$nl,
+		$_->[1],
+		$nl
+	    ) : "$final$nl$nl"
+	} @chunks).$end;
+    }
+    return "$hdr\r\n$before_chunks$data";
 }
 
 
