@@ -3,6 +3,7 @@ use warnings;
 package App::DubiousHTTP::Tests::Compressed;
 use App::DubiousHTTP::Tests::Common;
 use Compress::Raw::Zlib;
+use Compress::Raw::Lzma;
 
 SETUP( 
     'compressed',
@@ -25,11 +26,15 @@ DESC
 
     # these should be fine
     [ 'VALID: correct compressed requests' ],
-    [ VALID, 'ce:gzip;gzip' => 'content-encoding gzip, served gzipped'],
+    [ SHOULDBE_VALID, 'ce:gzip;gzip' => 'content-encoding gzip, served gzipped'],
     [ VALID, 'ce:gzip;gzip2p' => 'content-encoding gzip, served gzipped with 2 compressed blocks'],
-    [ VALID, 'ce:x-gzip;gzip' => 'content-encoding "x-gzip", served gzipped'],
-    [ VALID, 'ce:deflate;deflate' => 'content-encoding deflate, served with deflate'],
+    [ UNCOMMON_VALID, 'ce:x-gzip;gzip' => 'content-encoding "x-gzip", served gzipped'], # not IE11
+    [ SHOULDBE_VALID, 'ce:deflate;deflate' => 'content-encoding deflate, served with deflate'],
     [ VALID, 'ce:deflate;deflate2p' => 'content-encoding deflate, served with deflate with 2 compressed blocks'],
+    [ VALID, 'ce:deFLaTe;deflate' => 'content-encoding deflate mixed case, served with deflate'],
+
+    [ 'VALID: lzma (supported by at least Opera)' ],
+    [ UNCOMMON_VALID, 'ce:lzma;lzma1' => 'content-encoding lzma, lzma1 (lzma_alone) encoded'],
 
     # these might be strange/unsupported
     [ 'VALID: less common but valid requests' ],
@@ -38,6 +43,9 @@ DESC
     [ UNCOMMON_VALID, 'ce:nl-gzip;gzip' => 'content-encoding gzip but with continuation line, served gzipped'],
     [ UNCOMMON_VALID, 'ce:nl-deflate;deflate' => 'content-encoding deflate but with continuation line, served with deflate'],
     [ UNCOMMON_VALID, 'ce:nl-nl-deflate;deflate' => 'content-encoding deflate but with double continuation line, served with deflate'],
+    [ UNCOMMON_VALID, 'ce:deflate,;deflate' => 'content-encoding "deflate,", served with deflate'],
+    [ UNCOMMON_VALID, 'ce:deflate-nl-,;deflate' => 'content-encoding "deflate<nl> ,", served with deflate'],
+    [ UNCOMMON_VALID, 'ce:deflate-nl-,-nl-;deflate' => 'content-encoding "deflate<nl> ,<nl> ", served with deflate'],
 
     # These should be fine according to RFC, but are not supported in the browsers
     # Thus treat is as problem if they get supported.
@@ -52,6 +60,8 @@ DESC
     [ UNCOMMON_VALID, 'ce:gzip,gzip;gzip;gzip' => 'single content-encoding header "gzip,gzip", served twice gzipped'],
     [ UNCOMMON_VALID, 'ce:deflate;ce:deflate;deflate;deflate' => 'double content-encoding header deflate, compressed twice with deflate'],
     [ UNCOMMON_VALID, 'ce:deflate,deflate;deflate;deflate' => 'single content-encoding header "deflate,deflate", compressed twice with deflate'],
+    [ UNCOMMON_VALID, 'ce:deflate-nl-,-nl-deflate;deflate;deflate' => 'single content-encoding header "deflate<nl> ,<nl> deflate", compressed twice with deflate'],
+    [ UNCOMMON_VALID, 'ce:deflate-nl-,-nl-deflate-nl-;deflate;deflate' => 'single content-encoding header "deflate<nl> ,<nl> deflate<nl> ", compressed twice with deflate'],
 
     [ UNCOMMON_VALID, 'ce:gzip;ce:deflate;gzip;deflate' => 'content-encoding header for gzip and deflate, content compressed in this order'],
     [ UNCOMMON_VALID, 'ce:gzip,deflate;gzip;deflate' => 'single content-encoding "gzip,deflate", content compressed in this order'],
@@ -117,6 +127,10 @@ DESC
     [ INVALID, 'ce:deflate;gzip' => 'content-encoding deflate but served with gzip'],
     [ INVALID, 'ce:gzip;deflate' => 'content-encoding gzip but served with decode'],
     [ INVALID, 'ce:gzip;gzip-split' => 'content-encoding gzip, content split into 2 gzip parts concatenated'],
+    [ INVALID, 'ce:deflate' => 'content-encoding "deflate", not encoded'],
+    [ INVALID, 'ce:deflate,' => 'content-encoding "deflate,", not encoded'],
+    [ INVALID, 'ce:deflate-nl-,' => 'content-encoding "deflate<nl> ,", not encoded'],
+    [ INVALID, 'ce:deflate-nl-,-nl-' => 'content-encoding "deflate<nl> ,<nl> ", not encoded'],
 
     [ 'INVALID: invalid content-encodings should not be ignored' ],
     [ INVALID, 'ce:gzip_x' => 'content-encoding "gzip x", but not encoded' ],
@@ -149,6 +163,66 @@ DESC
     [ INVALID, 'lfonly-deflate' => 'Content-Encoding with only <LF> as line delimiter before, not served deflate' ],
     [ UNCOMMON_INVALID, 'lfonly-gzip;gzip' => 'Content-Encoding with only <LF> as line delimiter before, served gzip' ],
     [ INVALID, 'lfonly-gzip' => 'Content-Encoding with only <LF> as line delimiter before, not served gzip' ],
+
+    [ INVALID, 'ce:crdeflate;deflate' => 'Content-Encoding:<CR>deflate, served with deflate' ],
+    [ INVALID, 'ce:crdeflate' => 'Content-Encoding:<CR>deflate, not served with deflate' ],
+    [ INVALID, 'ce:cr-deflate;deflate' => 'Content-Encoding:<CR><space>deflate, served with deflate' ],
+    [ INVALID, 'ce:cr-deflate' => 'Content-Encoding:<CR><space>deflate, not served with deflate' ],
+    [ INVALID, 'ce:crgzip;gzip' => 'Content-Encoding:<CR>gzip, served with gzip' ],
+    [ INVALID, 'ce:crgzip' => 'Content-Encoding:<CR>gzip, not served with gzip' ],
+    [ INVALID, 'ce:cr-gzip;gzip' => 'Content-Encoding:<CR><space>gzip, served with gzip' ],
+    [ INVALID, 'ce:cr-gzip' => 'Content-Encoding:<CR><space>gzip, not served with gzip' ],
+
+    [ 'INVALID: slightly invalid gzip encodings' ],
+    [ INVALID,'ce:gzip;gzip;replace:0,2=1f8c', 'wrong gzip magic header'],
+    [ INVALID,'ce:gzip;gzip;replace:2,1=88', 'wrong compression method 88 instead of 08'],
+    [ UNCOMMON_VALID,'ce:gzip;gzip;replace:3,1|01', 'set flag FTEXT'],
+    [ INVALID,'ce:gzip;gzip;replace:3,1|02', 'set flag FHCRC without having CRC'],
+    [ INVALID,'ce:gzip;gzip;replace:3,1|02;replace:10,0=0000', 'set flag FHCRC and add CRC with 0'],
+    [ UNCOMMON_VALID,'ce:gzip;gzip;replace:3,1|04;replace:10,0=0000', 'set flag FEXTRA and extra part with XLEN 0'],
+    [ UNCOMMON_VALID,'ce:gzip;gzip;replace:3,1|04;replace:10,0=05004170010000', 'set flag FEXTRA and extra part with XLEN 5'],
+    [ UNCOMMON_VALID,'ce:gzip;gzip;replace:3,1|08;replace:10,0=2000', 'set flag FNAME and add short file name'],
+    [ UNCOMMON_VALID,'ce:gzip;gzip;replace:3,1|10;replace:10,0=2000', 'set flag FCOMMENT and add short comment'],
+    [ INVALID,'ce:gzip;gzip;replace:3,1|20', 'set flag reserved bit 5'],
+    [ INVALID,'ce:gzip;gzip;replace:3,1|40', 'set flag reserved bit 6'],
+    [ INVALID,'ce:gzip;gzip;replace:3,1|80', 'set flag reserved bit 7'],
+    [ INVALID,'ce:gzip;gzip;replace:-8,4^ffffffff', 'invalidate final checksum'],
+    [ INVALID,'ce:gzip;gzip;replace:-4,1^ff', 'invalidate length'],
+    [ INVALID,'ce:gzip;gzip;replace:-4,4=', 'remove length'],
+    [ INVALID,'ce:gzip;gzip;replace:-8,8=', 'remove checksum and length'],
+    [ INVALID,'ce:gzip;gzip;replace:-4,4=;clen+4', 'remove length but set content-length header to original size'],
+    [ INVALID,'ce:gzip;gzip;replace:-8,8=;clen+8', 'remove checksum and length but set content-length header to original size'],
+    [ INVALID,'ce:gzip;gzip;replace:-4,4=;noclen', 'remove length and close with eof without sending length'],
+    [ INVALID,'ce:gzip;gzip;replace:-8,8=;noclen', 'remove checksum and and close with eof without sending length'],
+    # and now hide the 'gzip' behind a \r so that some firewalls will use the
+    # heuristics of the antivirus which might be different from the the proxy
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|01', 'set flag FTEXT (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|02;replace:10,0=0000', 'set flag FHCRC and add CRC with 0 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|08;replace:10,0=2000', 'set flag FNAME and add short file name (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|10;replace:10,0=2000', 'set flag FCOMMENT and add short comment (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|04;replace:10,0=0000', 'set flag FEXTRA and extra part with XLEN 0 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|04;replace:10,0=05004170010000', 'set flag FEXTRA and extra part with XLEN 5 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|20', 'set flag reserved bit 5 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|40', 'set flag reserved bit 6 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:3,1|80', 'set flag reserved bit 7 (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:-8,4^ffffffff', 'invalidate final checksum (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:-4,1^ff', 'invalidate length (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:-4,4=', 'remove length (hide gzip with "content-encoding:\r gzip")'],
+    [ INVALID,'ce:cr-gzip;gzip;replace:-8,8=', 'remove checksum and length (hide gzip with "content-encoding:\r gzip")'],
+    # same game, but with Content-Encoding<space>: for other firewalls
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|01', 'set flag FTEXT (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|02;replace:10,0=0000', 'set flag FHCRC and add CRC with 0 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|08;replace:10,0=2000', 'set flag FNAME and add short file name (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|10;replace:10,0=2000', 'set flag FCOMMENT and add short comment (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|04;replace:10,0=0000', 'set flag FEXTRA and extra part with XLEN 0 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|04;replace:10,0=05004170010000', 'set flag FEXTRA and extra part with XLEN 5 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|20', 'set flag reserved bit 5 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|40', 'set flag reserved bit 6 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:3,1|80', 'set flag reserved bit 7 (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:-8,4^ffffffff', 'invalidate final checksum (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:-4,1^ff', 'invalidate length (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:-4,4=', 'remove length (hide gzip with "content-encoding : gzip")'],
+    [ INVALID,'ce-space-colon-gzip;gzip;replace:-8,8=', 'remove checksum and length (hide gzip with "content-encoding : gzip")'],
 );
 
 sub make_response {
@@ -156,16 +230,19 @@ sub make_response {
     return make_index_page() if $page eq '';
     my ($hdr,$data) = content($page,$spec) or die "unknown page $page";
     my $version = '1.1';
+    my $clen_extend;
     for (split(';',$spec)) {
-	if ( m{^(ce|te):(nl-(?:nl-)?)?(x_)?(x-gzip|x-deflate|gzip|deflate|xgzip|gzipx|foo|identity)(_x)?((?:,(?:deflate|gzip|identity))*)$} ) {
-	    $hdr .= $1 eq 'ce' ? 'Content-Encoding:':'Transfer-Encoding:';
-	    $hdr .= "\r\n " if $2;
-	    $hdr .= "\r\n " if $2 && $2 eq 'nl-nl-';
-	    $hdr .= "x " if $3;
-	    $hdr .= $4;
-	    $hdr .= $6 if $6;
-	    $hdr .= " x" if $5;
-	    $hdr .= "\r\n";
+	if ( my ($field,$v) = m{^(ce|te):(.*)$}i ) {
+	    my $changed;
+	    $changed++ if $v =~s{(?<=cr|lf|nl)-}{ }g;
+	    $changed++ if $v =~s{cr}{\r}g;
+	    $changed++ if $v =~s{lf}{\n}g;
+	    $changed++ if $v =~s{nl}{\r\n}g;
+	    $changed++ if $v =~s{_}{ }g;
+	    $v =~s{(?<!x)-}{}g;
+	    $hdr .= "Connection: close\r\n" if $changed;
+	    $hdr .= $field eq 'ce' ? 'Content-Encoding:':'Transfer-Encoding:';
+	    $hdr .= "$v\r\n";
 	} elsif ( m{^(?:(gzip)|deflate(-raw)?)(?:(\d+)p)?$} ) {
 	    my $zlib = Compress::Raw::Zlib::Deflate->new(
 		-WindowBits => $1 ? WANT_GZIP : $2 ? +MAX_WBITS() : -MAX_WBITS(),
@@ -195,6 +272,18 @@ sub make_response {
 		$newdata .= $out;
 	    }
 	    $data = $newdata;
+
+	} elsif (m{^(lzma[12]|xz)$}) {
+	    my ($lzma,$status) = 
+		$_ eq 'xz'    ? Compress::Raw::Lzma::EasyEncoder->new(AppendOutput => 1)  :
+		$_ eq 'lzma1' ? Compress::Raw::Lzma::AloneEncoder->new(AppendOutput => 1) :
+		Compress::Raw::Lzma::RawEncoder->new(AppendOutput =>1);
+	    $status == LZMA_OK or die "failed to create LZMA object";
+	    my $newdata = '';
+	    $lzma->code($data,$newdata) == LZMA_OK or die "failed to lzma encode data";
+	    $lzma->flush($newdata,LZMA_FINISH) == LZMA_STREAM_END or die "failed to close lzma stream";
+	    $data = $newdata;
+
 	} elsif (m{^ce-space-colon-(.*)}) {
 	    $hdr .= "Content-Encoding : $1\r\n";
 	} elsif (m{^ce-colon-colon-(.*)}) {
@@ -205,11 +294,35 @@ sub make_response {
 	    $hdr .= "Content-Encoding: $encoding\r\n";
 	} elsif ($_ eq 'ce:rfc2047-deflate') {
 	    $hdr .= "Content-Encoding: =?UTF-8?B?ZGVmbGF0ZQo=?=\r\n";
+	} elsif ( my ($off,$len,$op,$replacement) = m{replace:(-?\d+),(\d+)([=|^])(.*)}) {
+	    $replacement = pack('C*',map { hex($_) } $replacement=~m{(..)}g);
+	    if ($op eq '=') {
+		substr($data,$off,$len,$replacement);
+	    } elsif ($op eq '|') {
+		die "'$_' flags are already set" if (substr($data,$off,$len) & $replacement) eq $replacement;
+		substr($data,$off,$len) |= $replacement;
+	    } elsif ($op eq '^') {
+		substr($data,$off,$len) ^= $replacement;
+	    } else {
+		die "bad op=$op in '$_'"
+	    }
+	} elsif (m{^clen\+(\d+)$}) {
+	    $clen_extend = $1
+	} elsif ($_ eq 'noclen') {
+	    $clen_extend = -1;
 	} else {
 	    die $_
 	}
     }
-    $hdr .= "Content-length: ".length($data)."\r\n";
+
+    my $len = length($data);
+    if (!$clen_extend) {
+	$hdr = "Content-length: ".length($data)."\r\n".$hdr;
+    } elsif ($clen_extend<0) {
+	$hdr = "Connection: close\r\n$hdr";
+    } else {
+	$hdr = "Connection: close\r\nContent-length: ".(length($data)+$clen_extend)."\r\n".$hdr;
+    }
     return "HTTP/$version 200 ok\r\n$hdr\r\n$data";
 }
 
